{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

Author:       Franois Piette
Description:  TMWTrackDataSet keep track of TDataSet insert, delete and update
              and record that data into a TMWBuffer ready for sending to an
              application server.
Creation:     April 3, 1999
Version:      1.00
EMail:        http://www.overbyte.be        http://www.rtfm.be/fpiette
              francois.piette@overbyte.be   francois.piette@rtfm.be
              francois.piette@pophost.eunet.be
Support:      Use the mailing list midware@elists.org
              Follow "support" link at http://www.overbyte.be for subscription.
Legal issues: Copyright (C) 2002 by Franois PIETTE
              Rue de Grady 24, 4053 Embourg, Belgium. Fax: +32-4-365.74.56
              <francois.piette@overbyte.be>

              This software is provided 'as-is', without any express or
              implied warranty.  In no event will the author be held liable
              for any  damages arising from the use of this software.

              Permission is granted to anyone to use this software and or any
              derived or altered versions for any purpose, excluding commercial
              applications. You can use this software for personal use only.
              You may distribute it freely untouched.
              The following restrictions applies:

              1. The origin of this software must not be misrepresented, you
                 must not claim that you wrote the original software.

              2. If you use this software in a product, an acknowledgment in
                 the product documentation and displayed on screen is required.
                 The text must be: "This product is based on MidWare. Freeware
                 source code is available at http://www.overbyte.be."

              3. Altered source versions must be plainly marked as such, and
                 must not be misrepresented as being the original software.

              4. This notice may not be removed or altered from any source
                 distribution and must be added to the product documentation.


Updates:

 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
unit MWTrack;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  TrackDs, RFormat, Db;

type
    TMWTrackDataSet = class(TTrackDataSet)
    protected
        FData     : TMWBuffer;
        procedure   TriggerDataChanged; override;
        procedure   SaveData(TrackData : TTrackData);
        procedure   Notification(AComponent: TComponent;
                                 operation: TOperation); override;
    public
        constructor Create(AOwner : TComponent); override;
        destructor  Destroy; override;
        procedure   Clear; virtual;
        function    IsUploadNeeded : Boolean;
        procedure   LoadFromFile(FileName : String); virtual;
        procedure   SaveToFile(FileName : String); virtual;
        property    Data : TMWBuffer   read FData;
    end;

procedure Register;

implementation


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
constructor TMWTrackDataSet.Create(AOwner : TComponent);
begin
    inherited Create(AOwner);
    FData := TMWBuffer.Create(Self);
    Clear;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
destructor TMWTrackDataSet.Destroy;
begin
    if Assigned(FData) then begin
        FData.Destroy;
        FData := nil;
    end;
    inherited Destroy;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TMWTrackDataSet.Notification(
    AComponent: TComponent;
    operation: TOperation);
begin
    inherited Notification(AComponent, operation);
    if operation = opRemove then begin
        if AComponent = FData then
            FData := nil;
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TMWTrackDataSet.Clear;
begin
    if Assigned(FData) then
        FData.Rewrite;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
function TMWTrackDataSet.IsUploadNeeded : Boolean;
begin
    if not Assigned(FData) then begin
        Result := FALSE;
        Exit;
    end;
    FData.First;
    Result := not FData.Eof;
    if Result then
        Result := Length(FData.Fields[0]) > 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TMWTrackDataSet.SaveData(TrackData : TTrackData);
var
    Index        : Integer;
    Field        : TTrackDataField;
    Count        : Integer;
    IntegerData  : Integer;
    DateTimeData : TDateTime;
type
    PDateTimeRec = ^TDateTimeRec;
    PDateTime    = ^TDateTime;
begin
    Count  := TrackData.FieldCount - 1;
    for Index := 0 to Count do begin
        Field := TrackData.Fields[Index];
        case Field.DataType of
        ftInteger:
            begin
                IntegerData := PInteger(Field.Data)^;
                FData.AppendFields([IntegerData]);
            end;
        ftString:
            FData.AppendFields([PChar(Field.Data)]);
        ftBlob, ftMemo, ftGraphic:
            FData.AppendBinaryField(Field.Data, Field.DataSize, mwString);
        ftDate:
            begin
                DateTimeData := PDateTimeRec(Field.Data)^.Date - DateDelta;
                FData.AppendFields([DateTimeToMWFormat(DateTimeData)]);
            end;
        ftTime:
            begin
                DateTimeData := PDateTimeRec(Field.Data)^.Time / MSecsPerDay;
                FData.AppendFields([DateTimeToMWFormat(DateTimeData)]);
            end;
        ftDateTime :
            begin
                DateTimeData := (PDateTimeRec(Field.Data)^.DateTime /
                                 MSecsPerDay) - DateDelta;
                FData.AppendFields([DateTimeToMWFormat(DateTimeData)]);
            end;
        else
            raise ETrackDataSetException.Create('unsupported field type');
        end;
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TMWTrackDataSet.TriggerDataChanged;
begin
    if not Assigned(FData) then
        raise ETrackDataSetException.Create('FData not assigned');
    case Operation of
    tdsoInsert:
        begin
            FData.AppendBinaryField(PChar('INS'), 3, mwString);
            SaveData(DataAfter);
            FData.WriteRecordMarker;
        end;
    tdsoDelete:
        begin
            FData.AppendBinaryField(PChar('DEL'), 3, mwString);
            SaveData(DataBefore);
            FData.WriteRecordMarker;
        end;
    tdsoUpdate:
        begin
            if not DataBefore.Compare(DataAfter) then begin
                FData.AppendBinaryField(PChar('UPD'), 3, mwString);
                SaveData(DataBefore);
                FData.WriteRecordMarker;
                FData.AppendBinaryField(PChar('CHG'), 3, mwString);
                SaveData(DataAfter);
                FData.WriteRecordMarker;
            end;
        end;
    end;
    inherited TriggerDataChanged;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TMWTrackDataSet.LoadFromFile(FileName : String);
begin
    if Assigned(FData) then
        FData.LoadFromFile(FileName);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TMWTrackDataSet.SaveToFile(FileName : String);
begin
    if Assigned(FData) then
        FData.SaveToFile(FileName);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure Register;
begin
    RegisterComponents('FPiette', [TMWTrackDataSet]);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

end.
